"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.sqitchDeploy = exports.runWatcher = exports.runCmd = exports.restartContainer = exports.resetDb = exports.notEmptyString = exports.fileExists = exports.dockerContainers = exports.dirExists = exports.checkPostgresConnection = exports.checkOpenrestyInitiated = exports.checkMigrationsInitiated = exports.checkIsAppDir = void 0;

var _child_process = _interopRequireDefault(require("child_process"));

var _fs = _interopRequireDefault(require("fs"));

var _os = _interopRequireDefault(require("os"));

var _chokidar = _interopRequireDefault(require("chokidar"));

var _string_decoder = require("string_decoder");

var _colors = _interopRequireDefault(require("colors"));

var _env = require("./env.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const runCmd = (cmd, params, options = {}, silent = false, exit_on_error = false, use_async = false, stdio) => {
  if (_env.USE_DOCKER_IMAGE && [_env.PSQL_CMD, _env.SQITCH_CMD, _env.PG_DUMP_CMD, _env.PG_DUMPALL_CMD, _env.JAVA_CMD, _env.MIGRA_CMD].indexOf(cmd) !== -1) {
    //alter the command to run in docker
    let app_dir_regexp = new RegExp(_env.APP_DIR, 'g');
    let w = options && options.cwd ? options.cwd.replace(_env.APP_DIR, _env.DOCKER_APP_DIR) : _env.DOCKER_APP_DIR;
    let e = options && options.env ? Object.keys(options.env).reduce(function (acc, key) {
      acc.push('-e');
      acc.push(key + '=' + options.env[key].replace(app_dir_regexp, _env.DOCKER_APP_DIR));
      return acc;
    }, []) : [];
    let u = [],
        hostAppDir = _env.APP_DIR;

    if (_env.PLATFORM == 'linux') {
      let {
        uid,
        gid
      } = _os.default.userInfo();

      u = ['-u', `${uid}:${gid}`, '--env', "USERNAME=root"];
    }

    if (_env.PLATFORM == 'win32' && _env.DOCKER_HOST_OS != "Docker for Windows") {
      // docker is running in virtualbox VM so we try to adjust the path to match the shared path
      hostAppDir = '/' + hostAppDir.replace(/([A-Z]):/, function (s) {
        return s.toLocaleLowerCase().replace(':', '');
      }).replace(/\\/g, '/');
    }

    let ef = fileExists(`.env`) ? ['--env-file', `${_env.APP_DIR}/.env`] : [];
    let p = ['run', '--net', 'host', '--rm', '-w', w, '-v', `${hostAppDir}:${_env.DOCKER_APP_DIR}`].concat(e).concat(u).concat(ef).concat([_env.DOCKER_IMAGE, cmd]).concat(params.map(v => v.replace(app_dir_regexp, _env.DOCKER_APP_DIR)));
    cmd = 'docker';
    params = p;
  }

  if (!(silent || use_async)) {
    options.stdio = stdio || ['ignore', 1, 2];
  }

  let spawn = use_async ? _child_process.default.spawn : _child_process.default.spawnSync; //console.log(cmd, params, options);

  let pr = spawn(cmd, params, options);

  if (!use_async && exit_on_error && pr.status != 0) {
    process.exit(pr.status);
  }

  return pr;
};

exports.runCmd = runCmd;

const checkPostgresConnection = (connection_string, pass, exit_on_error = true) => {
  console.log(`Checking PostgreSQL connection info (${connection_string})`);
  var env = Object.create(process.env);
  env.PGPASSWORD = pass;
  const random_number = Math.floor(Math.random() * 100).toString();
  const params = ['--quiet', '--tuples-only', '-c', `SELECT ${random_number}`, connection_string];
  let result = runCmd(_env.PSQL_CMD, params, {
    env: env
  }, true, false).stdout.toString().replace(/\s/g, '');

  if (random_number !== result) {
    console.log(" ");
    console.log(`Could not verify the connection to PostgreSQL`);

    if (exit_on_error) {
      process.exit(1);
    } else {
      return false;
    }
  }

  return true;
};

exports.checkPostgresConnection = checkPostgresConnection;

const sqitchDeploy = url => runCmd(_env.SQITCH_CMD, ["deploy", url], {
  cwd: _env.MIGRATIONS_DIR
}, false, true);

exports.sqitchDeploy = sqitchDeploy;

const fileExists = path => _fs.default.existsSync(path) && _fs.default.statSync(path).isFile();

exports.fileExists = fileExists;

const dirExists = path => _fs.default.existsSync(path) && _fs.default.statSync(path).isDirectory(); // options.key from commander returns bool if a value is not specified(e.g. subzero cloud login -u, options.username gives true), so make sure is a string


exports.dirExists = dirExists;

const notEmptyString = s => typeof s == 'string' && s.trim().length;

exports.notEmptyString = notEmptyString;

const checkIsAppDir = () => {
  if (!fileExists(`.env`)) {
    console.log("Error: ".red + ".env file does not exist");
    console.log("Please run this command in a directory that contains a subzero project.");
    process.exit(0);
  }
};

exports.checkIsAppDir = checkIsAppDir;

const checkMigrationsInitiated = () => {
  if (!(dirExists(_env.MIGRATIONS_DIR) && fileExists(`${_env.MIGRATIONS_DIR}/sqitch.plan`))) {
    console.log("Error: ".red + "database migrations not initiated");
    console.log("Please run `subzero migrations init` before trying to deploy the code");
    process.exit(0);
  }
};

exports.checkMigrationsInitiated = checkMigrationsInitiated;

const checkOpenrestyInitiated = () => {
  if (!(dirExists(_env.OPENRESTY_DIR) && fileExists(`${_env.OPENRESTY_DIR}/Dockerfile`))) {
    console.log("Error: ".red + "Dockerfile for custom OpenResty image missing");
    console.log(`Please create ${_env.OPENRESTY_DIR}/Dockerfile`);
    process.exit(0);
  }
};

exports.checkOpenrestyInitiated = checkOpenrestyInitiated;
const decoder = new _string_decoder.StringDecoder('utf8');

const resetDb = (containers, logger) => {
  logger.log('starting db reset');
  const recreatedb = runSql(['postgres', '--quiet', '-c', `SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = '${_env.DB_NAME}';`, '-c', `DROP DATABASE if exists ${_env.DB_NAME};`, '-c', `CREATE DATABASE ${_env.DB_NAME};`], false);
  let err = decoder.write(recreatedb.stderr).trim();
  if (err.length > 0) logger.log(err);
  const psql = runSql([_env.DB_NAME, '-f', _env.DB_DIR + _env.SQL_INIT_FILE], false);
  err = decoder.write(psql.stderr).trim();
  if (err.length > 0) logger.log(err);

  if (psql.status == 0) {
    //support old style starter kit
    if (containers['postgrest']) {
      sendHUP(containers['postgrest'].name, 'SIGUSR1');
    }

    if (containers['openresty']) {
      sendHUP(containers['openresty'].name);
    }

    logger.log('db reset done');
  } else {
    logger.log('db reset failed');
  }

  return psql;
};

exports.resetDb = resetDb;

const sendHUP = (containerName, signal = 'HUP') => _child_process.default.spawn('docker', ['kill', '-s', signal, containerName]);

const restartContainer = (containerName, callback) => _child_process.default.spawn('docker', ['restart', containerName]).on('close', callback);

exports.restartContainer = restartContainer;

const runSql = (commands, use_async = true) => {
  const connectParams = ['-U', _env.SUPER_USER, '-h', 'localhost', '--set', 'DIR=' + _env.DB_DIR + '/', '--set', 'ON_ERROR_STOP=1'];
  var env = Object.create(process.env);
  env.PGPASSWORD = _env.SUPER_USER_PASSWORD;
  env.DIR = _env.DB_DIR;
  const stdio = use_async ? undefined : ['ignore', 'pipe', 'pipe'];
  return runCmd(_env.PSQL_CMD, connectParams.concat(commands), {
    env: env
  }, undefined, undefined, use_async, stdio);
};

const runWatcher = (containers, logger, watcherReadyCb, reloadStartCb, reloadEndCb) => {
  return _chokidar.default.watch(_env.WATCH_PATTERNS, {
    ignored: _env.APP_DIR + '/**/tests/**'
  }).on('change', path => {
    const relPath = path.replace(_env.APP_DIR, '.');
    reloadStartCb(relPath);

    if (path.endsWith('.sql')) {
      const result = resetDb(containers, logger);
      reloadEndCb(result.status);
    } else {
      if (containers['openresty']) {
        sendHUP(containers['openresty'].name).on('close', function () {
          reloadEndCb(0);
        });
      } else if (containers['subzero']) {
        restartContainer(containers['subzero'].name, function () {
          reloadEndCb(0);
        });
      } else {
        reloadEndCb(0);
      }
    }
  }).on('ready', watcherReadyCb);
};

exports.runWatcher = runWatcher;

const dockerContainers = () => {
  const filters_params = _child_process.default.execSync(`docker-compose ps -q 2> /dev/null | cut -c 1-12`).toString('utf8').trim().split("\n").map(id => '--filter id=' + id);

  const containers = _child_process.default.execSync(`docker ps -a --format "{{.Names}}" ` + filters_params.join(' ')).toString('utf8').trim().split("\n");

  const result = containers.reduce((acc, containerName) => {
    let key = containerName.replace(_env.COMPOSE_PROJECT_NAME, '').replace('1', '').replace(/_/g, '');
    if (key == "") return acc;
    acc[key] = {
      name: containerName,
      title: containerName.replace(_env.COMPOSE_PROJECT_NAME + '_', '').replace(/_1$/, '')
    };
    return acc;
  }, {});
  return result;
};

exports.dockerContainers = dockerContainers;