#!/usr/bin/env node
"use strict";

var _commander = _interopRequireDefault(require("commander"));

var _child_process = _interopRequireDefault(require("child_process"));

var _fs = _interopRequireDefault(require("fs"));

var _sleep = _interopRequireDefault(require("sleep"));

var _rimraf = _interopRequireDefault(require("rimraf"));

var _common = require("./common.js");

var _net = require("net");

var _env = require("./env.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const TMP_DIR = `${_env.MIGRATIONS_DIR}/tmp`;
const INITIAL_FILE_NAME = "initial";
const SQITCH_CONF = `${_env.MIGRATIONS_DIR}/sqitch.conf`;
const INIT_SQL_FILENAME = '0000000000-setup.sql';

const getFreePort = async () => {
  return new Promise((resolve, reject) => {
    const server = (0, _net.createServer)();
    server.on('error', reject);
    server.on('listening', () => (port => server.close(() => resolve(port)))(server.address().port));
    server.listen(0);
  });
};

const processIncludeFiles = (migrationNumber, migrationName, withFilesBeginning = '', withFilesEnd = '') => {
  withFilesBeginning = withFilesBeginning ? withFilesBeginning.trim() : '';
  withFilesEnd = withFilesEnd ? withFilesEnd.trim() : '';
  withFilesBeginning = withFilesBeginning === '' ? [] : withFilesBeginning.split(',').map(f => f.trim());
  withFilesEnd = withFilesEnd === '' ? [] : withFilesEnd.split(',').map(f => f.trim());
  withFilesBeginning = withFilesBeginning.map(f => {
    return f && !_fs.default.existsSync(f) && _fs.default.existsSync(`${_env.DB_DIR}/${f}`) ? `${_env.DB_DIR}/${f}` : f;
  });
  withFilesEnd = withFilesEnd.map(f => {
    return f && !_fs.default.existsSync(f) && _fs.default.existsSync(`${_env.DB_DIR}/${f}`) ? `${_env.DB_DIR}/${f}` : f;
  });
  let allFilesFound = true;
  withFilesBeginning.forEach(f => {
    if (!_fs.default.existsSync(f)) {
      console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", f);
      allFilesFound = false;
    }
  });
  withFilesEnd.forEach(f => {
    if (!_fs.default.existsSync(f)) {
      console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", f);
      allFilesFound = false;
    }
  });

  if (!allFilesFound) {
    process.exit(-1);
  }

  return {
    withFilesBeginning: withFilesBeginning,
    withFilesEnd: withFilesEnd,
    withFilesBeginningMigrationName: withFilesBeginning.map((_f, i) => `${migrationNumber}-${migrationName}.${i}.beginning`),
    withFilesEndMigrationName: withFilesEnd.map((_f, i) => `${migrationNumber}-${migrationName}.${i}.end`)
  };
};

const initMigrations = async (debug, dbDockerImage, dryRun, diffTool, includeRoles, includePrivileges, sWithFilesBeginning = '', sWithFilesEnd = '') => {
  if (_fs.default.existsSync(_env.MIGRATIONS_DIR)) {
    console.log(`Migrations directory already exists: ${_env.MIGRATIONS_DIR}`);
    process.exit(-1);
  }

  let tempDevDbUri = null,
      tempProdDbUri = null,
      tempProdDbcontainerName = null;
  const name = INITIAL_FILE_NAME,
        DIFF_TOOL = diffTool ? diffTool : _env.SQL_DIFF_TOOL,
        migrationNumber = getMigrationNumber(),
        tmpDevSql = `${TMP_DIR}/${migrationNumber}-dev-${name}.sql`,
        tmpProdSql = `${TMP_DIR}/${migrationNumber}-prod-${name}.sql`,
        tmpDeploySql = `${TMP_DIR}/${migrationNumber}-deploy-${name}.sql`,
        tmpRevertSql = `${TMP_DIR}/${migrationNumber}-revert-${name}.sql`,
        {
    withFilesBeginning: withFilesBeginning,
    withFilesEnd: withFilesEnd,
    withFilesBeginningMigrationName: withFilesBeginningMigrationName,
    withFilesEndMigrationName: withFilesEndMigrationName
  } = processIncludeFiles(migrationNumber, name, sWithFilesBeginning, sWithFilesEnd);

  _fs.default.mkdirSync(_env.MIGRATIONS_DIR);

  _fs.default.mkdirSync(TMP_DIR);

  initSqitch();
  const {
    containerName: devDbcontainerName,
    dbUri: devDbUri,
    initSqlFile: devInitSqlFile
  } = await getTempPostgres(_env.DB_DIR, debug, dbDockerImage);
  const devDbContianerReady = waitForDbContainer(devDbcontainerName, devInitSqlFile, debug);

  if (!devDbContianerReady) {
    stopContainer(devDbcontainerName);

    _rimraf.default.sync(_env.MIGRATIONS_DIR);

    process.exit(-1);
  }

  tempDevDbUri = devDbUri;
  dumpSchema(devDbUri, tmpDevSql, includeRoles, includePrivileges, withFilesBeginningMigrationName, withFilesEndMigrationName);

  _fs.default.closeSync(_fs.default.openSync(tmpProdSql, 'w'));

  if (DIFF_TOOL !== 'apgdiff') {
    const {
      containerName: prodDbcontainerName,
      dbUri: prodDbUri,
      initSqlFile: prodInitSqlFile
    } = await getTempPostgres(`${_env.MIGRATIONS_DIR}/deploy`, debug, dbDockerImage);
    const prodDbContianerReady = waitForDbContainer(prodDbcontainerName, prodInitSqlFile, debug);

    if (!prodDbContianerReady) {
      stopContainer(devDbcontainerName);
      stopContainer(prodDbcontainerName);

      _rimraf.default.sync(_env.MIGRATIONS_DIR);

      process.exit(-1);
    }

    tempProdDbUri = prodDbUri;
    tempProdDbcontainerName = prodDbcontainerName;
  } // in this case, the database dump is the initial migration
  // we only generate the revert migration


  switch (DIFF_TOOL) {
    case 'apgdiff':
      //apgdiffToFile(tmpProdSql, tmpDevSql, tmpDeploySql);
      apgdiffToFile(tmpDevSql, tmpProdSql, tmpRevertSql, [], []);
      break;

    case 'migra':
      //migraToFile(tempProdDbUri, tempDevDbUri, tmpDeploySql, includeRoles, includePrivileges);
      migraToFile(tempDevDbUri, tempProdDbUri, tmpRevertSql, false, false, [], []);
      break;
    // case 'sqlworkbench':
    //   sqlworkbenchFile(tempDevDbUri, tempProdDbUri, tmpDeploySql, includeRoles, includePrivileges);
    //   sqlworkbenchFile(tempProdDbUri, tempDevDbUri, tmpRevertSql, includeRoles, includePrivileges);
    //   break;
  }

  stopContainer(devDbcontainerName);

  if (tempProdDbUri) {
    stopContainer(tempProdDbcontainerName);
  }

  addSqitchMigration(`${migrationNumber}-${name}`);
  withFilesBeginning.forEach((f, i) => {
    _fs.default.copyFileSync(f, `${_env.MIGRATIONS_DIR}/deploy/${withFilesBeginningMigrationName[i]}`);
  });
  withFilesEnd.forEach((f, i) => {
    _fs.default.copyFileSync(f, `${_env.MIGRATIONS_DIR}/deploy/${withFilesEndMigrationName[i]}`);
  });
  console.log(`Copying ${tmpDevSql.replace(_env.APP_DIR, '')} to ${_env.MIGRATIONS_DIR.replace(_env.APP_DIR, '')}/deploy/${migrationNumber}-${name}.sql`);

  _fs.default.copyFileSync(tmpDevSql, `${_env.MIGRATIONS_DIR}/deploy/${migrationNumber}-${name}.sql`);

  if (_fs.default.existsSync(tmpRevertSql)) {
    _fs.default.copyFileSync(tmpRevertSql, `${_env.MIGRATIONS_DIR}/revert/${migrationNumber}-${name}.sql`);
  } else {
    console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", `${migrationNumber}-revert-${name}.sql`);
  }

  if (!debug) {
    if (_fs.default.existsSync(tmpDevSql)) {
      _fs.default.unlinkSync(tmpDevSql);
    }

    if (_fs.default.existsSync(tmpProdSql)) {
      _fs.default.unlinkSync(tmpProdSql);
    }

    if (_fs.default.existsSync(tmpRevertSql)) {
      _fs.default.unlinkSync(tmpRevertSql);
    }

    if (_fs.default.existsSync(tmpDeploySql)) {
      _fs.default.unlinkSync(tmpDeploySql);
    }
  }

  if (dryRun) {
    //we just delete the dir
    _rimraf.default.sync(_env.MIGRATIONS_DIR);
  }
};

const addMigration = async (name, note, diff, dryRun, debug, dbUri, dbDockerImage, diffTool, includePrivileges, sWithFilesBeginning = '', sWithFilesEnd = '') => {
  if (!_fs.default.existsSync(SQITCH_CONF) || !_fs.default.statSync(SQITCH_CONF).isFile()) {
    console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", SQITCH_CONF);
    console.log("You need to run 'subzero migrations init'");
    process.exit(-1);
  }

  if (dbUri) {
    (0, _common.checkPostgresConnection)(dbUri, '');
  }

  const migrationNumber = getMigrationNumber(),
        tmpDevSql = `${TMP_DIR}/${migrationNumber}-dev-${name}.sql`,
        tmpProdSql = `${TMP_DIR}/${migrationNumber}-prod-${name}.sql`,
        tmpDeploySql = `${TMP_DIR}/${migrationNumber}-deploy-${name}.sql`,
        tmpRevertSql = `${TMP_DIR}/${migrationNumber}-revert-${name}.sql`,
        DIFF_TOOL = diffTool ? diffTool : _env.SQL_DIFF_TOOL,
        {
    withFilesBeginning: withFilesBeginning,
    withFilesEnd: withFilesEnd,
    withFilesBeginningMigrationName: withFilesBeginningMigrationName,
    withFilesEndMigrationName: withFilesEndMigrationName
  } = processIncludeFiles(migrationNumber, name, sWithFilesBeginning, sWithFilesEnd);

  if (diff) {
    if (!_fs.default.existsSync(TMP_DIR)) _fs.default.mkdirSync(TMP_DIR);
    let tempDevDbUri = null,
        tempProdDbUri = null,
        tempProdDbcontainerName = null;
    const {
      containerName: devDbcontainerName,
      dbUri: devDbUri,
      initSqlFile: devInitSqlFile
    } = await getTempPostgres(_env.DB_DIR, debug, dbDockerImage);
    const devDbContianerReady = waitForDbContainer(devDbcontainerName, devInitSqlFile, debug);

    if (!devDbContianerReady) {
      stopContainer(devDbcontainerName);
      process.exit(-1);
    }

    tempDevDbUri = devDbUri;

    if (DIFF_TOOL == 'apgdiff') {
      dumpSchema(devDbUri, tmpDevSql, false, includePrivileges, [], []);
    }

    if (dbUri) {
      tempProdDbUri = dbUri;

      if (DIFF_TOOL == 'apgdiff') {
        dumpSchema(dbUri, tmpProdSql, false, includePrivileges, [], []);
      }
    } else {
      const {
        containerName: prodDbcontainerName,
        dbUri: prodDbUri,
        initSqlFile: prodInitSqlFile
      } = await getTempPostgres(`${_env.MIGRATIONS_DIR}/deploy`, debug, dbDockerImage);
      const prodDbContianerReady = waitForDbContainer(prodDbcontainerName, prodInitSqlFile, debug);

      if (!prodDbContianerReady) {
        stopContainer(devDbcontainerName);
        stopContainer(prodDbcontainerName);
        process.exit(-1);
      }

      if (DIFF_TOOL == 'apgdiff') {
        dumpSchema(prodDbUri, tmpProdSql, false, includePrivileges, [], []);
      }

      tempProdDbUri = prodDbUri;
      tempProdDbcontainerName = prodDbcontainerName;
    }

    switch (DIFF_TOOL) {
      case 'apgdiff':
        apgdiffToFile(tmpProdSql, tmpDevSql, tmpDeploySql, withFilesBeginningMigrationName, withFilesEndMigrationName);
        apgdiffToFile(tmpDevSql, tmpProdSql, tmpRevertSql, [], []);
        break;

      case 'migra':
        migraToFile(tempProdDbUri, tempDevDbUri, tmpDeploySql, false, includePrivileges, withFilesBeginningMigrationName, withFilesEndMigrationName);
        migraToFile(tempDevDbUri, tempProdDbUri, tmpRevertSql, false, includePrivileges, [], []);
        break;
      // case 'sqlworkbench':
      //   sqlworkbenchFile(tempDevDbUri, tempProdDbUri, tmpDeploySql, includeRoles, includePrivileges);
      //   sqlworkbenchFile(tempProdDbUri, tempDevDbUri, tmpRevertSql, includeRoles, includePrivileges);
      //   break;
    }

    stopContainer(devDbcontainerName);

    if (!dbUri) {
      stopContainer(tempProdDbcontainerName);
    }
  }

  if (!dryRun) {
    addSqitchMigration(`${migrationNumber}-${name}`, note);

    if (diff) {
      withFilesBeginning.forEach((f, i) => {
        _fs.default.copyFileSync(f, `${_env.MIGRATIONS_DIR}/deploy/${withFilesBeginningMigrationName[i]}`);
      });
      withFilesEnd.forEach((f, i) => {
        _fs.default.copyFileSync(f, `${_env.MIGRATIONS_DIR}/deploy/${withFilesEndMigrationName[i]}`);
      });

      if (_fs.default.existsSync(tmpDeploySql)) {
        _fs.default.copyFileSync(tmpDeploySql, `${_env.MIGRATIONS_DIR}/deploy/${migrationNumber}-${name}.sql`);
      } else {
        console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", `${migrationNumber}-deploy-${name}.sql`);
      }

      if (_fs.default.existsSync(tmpRevertSql)) {
        _fs.default.copyFileSync(tmpRevertSql, `${_env.MIGRATIONS_DIR}/revert/${migrationNumber}-${name}.sql`);
      } else {
        console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", `${migrationNumber}-revert-${name}.sql`);
      }

      console.log(`\x1b[31mATTENTION:\x1b[0m Make sure you check deploy/${migrationNumber}-${name}.sql for correctness, statement order is not handled!`);
    } else {
      console.log('Creating empty migration');
    }
  } else {
    console.log("\nCurrent migration DDL");
    console.log("=====================================================");

    if (_fs.default.existsSync(tmpDeploySql)) {
      console.log(_fs.default.readFileSync(tmpDeploySql, 'utf-8'));
    }
  }

  if (!debug) {
    if (_fs.default.existsSync(tmpDevSql)) {
      _fs.default.unlinkSync(tmpDevSql);
    }

    if (_fs.default.existsSync(tmpProdSql)) {
      _fs.default.unlinkSync(tmpProdSql);
    }

    if (_fs.default.existsSync(tmpRevertSql)) {
      _fs.default.unlinkSync(tmpRevertSql);
    }

    if (_fs.default.existsSync(tmpDeploySql)) {
      _fs.default.unlinkSync(tmpDeploySql);
    }
  }
};

const getMigrationNumber = () => new Date().toISOString().slice(0, 19).replace(/[^0-9]/g, '');

const stopContainer = name => {
  (0, _common.runCmd)("docker", ["stop", name], undefined, true);
  (0, _common.runCmd)("docker", ["rm", name], undefined, true);
};

const writeInitSql = file => {
  _fs.default.writeFileSync(file, ['-- custom setup sql', `create role "${_env.DB_USER}" with login password '${_env.DB_PASS}';`].join("\n"));
};

const getTempPostgres = async (sqlDir, debug, dbDockerImage) => {
  const name = 'temp_postgres_' + Math.random().toString(36).substr(2, 5);
  const port = await getFreePort();
  const initSqlFile = `${sqlDir}/${INIT_SQL_FILENAME}`;
  writeInitSql(initSqlFile);
  let pgVersion = getPgVersion(_env.DEV_DB_URI);
  let pgDockerImage = dbDockerImage ? dbDockerImage : `postgres:${pgVersion}`;
  console.log(`Starting temporary PostgreSQL database with the version matching your dev database (${pgDockerImage})`);
  console.log(`on port ${port} with SQL from ${sqlDir.replace(_env.APP_DIR, '')}`);
  const cmd = 'docker',
        params = ["run", "-d", "--name", name, "-p", port + ":5432", //# env vars specific to postgres image used on first boot
  "-e", `POSTGRES_USER=${_env.SUPER_USER}`, "-e", `POSTGRES_PASSWORD=${_env.SUPER_USER_PASSWORD}`, "-e", `POSTGRES_DB=${_env.DB_NAME}`, //# env vars useful for our sql scripts
  "-e", `SUPER_USER=${_env.SUPER_USER}`, "-e", `SUPER_USER_PASSWORD=${_env.SUPER_USER_PASSWORD}`, "-e", `DB_NAME=${_env.DB_NAME}`, "-e", `DB_USER=${_env.DB_USER}`, "-e", `DB_PASS=${_env.DB_PASS}`, "-e", `DB_ANON_ROLE=${process.env.DB_ANON_ROLE}`, "-e", `DEVELOPMENT=${process.env.DEVELOPMENT}`, "-e", `JWT_SECRET=${process.env.JWT_SECRET}`, "-v", `${sqlDir}:/docker-entrypoint-initdb.d`, pgDockerImage];

  if (debug) {
    console.log("Starting temp contianer with:");
    console.log(`${cmd} ${params.join(' ')}`);
  }

  (0, _common.runCmd)(cmd, params);
  return {
    containerName: name,
    dbUri: `postgresql://${_env.SUPER_USER}:${_env.SUPER_USER_PASSWORD}@${_env.LOCALHOST}:${port}/${_env.DB_NAME}`,
    initSqlFile: initSqlFile
  };
};

const waitForDbContainer = (name, initSqlFile, debug) => {
  console.log(`Waiting for ${name} container to load`);
  let finishedLoading = false;
  let timestamp = 0;
  let iterations = 0;
  const maxIterations = 60;

  while (!finishedLoading) {
    iterations = iterations + 1;

    if (iterations > maxIterations) {
      break;
    }

    let p0 = _child_process.default.spawnSync('docker', ['ps', '-a', '--filter', `name=${name}`, '--format', '"{{.Status}}"']);

    let output = p0.output.map(l => l ? l.toString() : '').join(' ');

    if (output.indexOf('Exited') !== -1) {
      //container has exited with an error, no point in waiting
      break;
    }

    let p = _child_process.default.spawnSync('docker', ['logs', '--since', timestamp, name]);

    timestamp = Math.floor(new Date() / 1000);
    p.output.forEach(v => {
      if (v) {
        if (v.toString().indexOf('PostgreSQL init process complete; ready for start up.') !== -1) {
          finishedLoading = true;
        }
      }
    });

    _sleep.default.sleep(1);
  }

  _sleep.default.sleep(1);

  if (_fs.default.existsSync(initSqlFile)) {
    _fs.default.unlinkSync(initSqlFile);
  }

  ;

  if (finishedLoading) {
    console.log('PostgreSQL init process complete; ready for start up.');

    if (debug) {
      (0, _common.runCmd)('docker', ['logs', name]);
    }
  } else {
    (0, _common.runCmd)('docker', ['logs', '--tail', '30', name]);
    console.log('\x1b[31mError:\x1b[0m Gave up on waiting for db. The last 30 lines of log are above.');

    if (_fs.default.existsSync(initSqlFile)) {
      _fs.default.unlinkSync(initSqlFile);
    }

    ;
  }

  return finishedLoading;
};

const initSqitch = () => (0, _common.runCmd)(_env.SQITCH_CMD, ["init", _env.DB_NAME, "--engine", "pg"], {
  cwd: _env.MIGRATIONS_DIR
});

const addSqitchMigration = (name, note) => (0, _common.runCmd)(_env.SQITCH_CMD, ["add", name, "-n", note || `Add ${name} migration`], {
  cwd: _env.MIGRATIONS_DIR
});

const dumpRoles = (dbUri, file) => {
  console.log(`Writing database roles dump to ${file.replace(_env.APP_DIR, '')}`);
  const replace_superuser = new RegExp(`GRANT ([a-z0-9_-]+) TO ${_env.SUPER_USER}`, "gi");
  (0, _common.runCmd)(_env.PG_DUMPALL_CMD, ['-f', `${file}`, '--roles-only', '-d', dbUri]); // we need some small postprocessing 

  let data = _fs.default.readFileSync(`${file}`, 'utf-8').split("\n").filter(ln => _env.IGNORE_ROLES.map(r => ln.indexOf('ROLE ' + r)).every(p => p == -1)) //filter out line referring to ignored roles
  .map(ln => ln.replace(` GRANTED BY ${_env.SUPER_USER}`, '')) //remove unwanted string
  .filter(ln => ln.indexOf('ALTER ROLE') == -1) //RDS does not allow this
  .map(ln => ln.replace(replace_superuser, 'GRANT $1 TO current_user')).join("\n");

  _fs.default.writeFileSync(file, data, 'utf-8');
};

const dumpSchema = (dbUri, file, includeRoles, includePrivileges, withFilesBeginning = [], withFilesEnd = []) => {
  console.log(`Writing database dump to ${file.replace(_env.APP_DIR, '')}`);

  if (includeRoles) {
    dumpRoles(dbUri, `${file}.roles`);
  }

  let params = [dbUri, '-f', `${file}.schema`, '--schema-only'];

  if (!includePrivileges) {
    params.push('--no-privileges');
  }

  (0, _common.runCmd)(_env.PG_DUMP_CMD, params);
  let data = ['-- generated with subzero-cli (https://github.com/subzerocloud/subzero-cli)', 'BEGIN;', withFilesBeginning.map(f => `\\ir ${f}`).join("\n"), !includeRoles ? '' : _fs.default.readFileSync(`${file}.roles`, 'utf-8'), _fs.default.readFileSync(`${file}.schema`, 'utf-8').split("\n").filter(ln => ln.indexOf('COMMENT ON EXTENSION') == -1) //RDS does not allow this
  .filter(ln => ln.indexOf(`OWNER TO ${_env.SUPER_USER};`) == -1) //don't keep owner info when the owner is privileges
  .join("\n"), withFilesEnd.map(f => `\\ir ${f}`).join("\n"), 'COMMIT;'];

  _fs.default.writeFileSync(file, data.join("\n"), 'utf-8');

  if (includeRoles) {
    _fs.default.unlinkSync(`${file}.roles`);
  }

  _fs.default.unlinkSync(`${file}.schema`);
};

const apgdiffToFile = (file1, file2, destFile, withFilesBeginning = [], withFilesEnd = []) => {
  let cmd = _env.JAVA_CMD;
  let params = ['-jar', _env.APGDIFF_JAR_PATH, file1, file2];
  let options = {};

  if (_env.USE_DOCKER_IMAGE && [_env.SQITCH_CMD, _env.PG_DUMP_CMD, _env.PG_DUMPALL_CMD, _env.JAVA_CMD].indexOf(cmd) !== -1) {
    //alter the command to run in docker
    let w = options && options.cwd ? options.cwd.replace(_env.APP_DIR, _env.DOCKER_APP_DIR) : _env.DOCKER_APP_DIR;
    params = ['run', '--rm', '-w', w, '-v', `${_env.APP_DIR}:${_env.DOCKER_APP_DIR}`, _env.DOCKER_IMAGE, cmd].concat(params.map(p => p.replace(_env.APP_DIR, _env.DOCKER_APP_DIR)));
    cmd = 'docker';
  }

  console.log(`Diffing ${file1.replace(_env.APP_DIR, '')} and ${file2.replace(_env.APP_DIR, '')}`);
  console.log(`Writing the result to ${destFile.replace(_env.APP_DIR, '')}`);

  let p = _child_process.default.spawnSync(cmd, params, options);

  if (p.stdout) {
    _fs.default.writeFileSync(destFile, ['-- generated with subzero-cli (https://github.com/subzerocloud/subzero-cli)', 'BEGIN;', withFilesBeginning.map(f => `\\ir ${f}`).join("\n"), p.stdout.toString(), withFilesEnd.map(f => `\\ir ${f}`).join("\n"), 'COMMIT;'].join("\n"));
  }

  if (p.stderr.toString()) console.log(p.stderr.toString());
}; // const diffRoles = (dburl_from, dburl_target, destFile) => {
//   dumpRoles(dburl_from, `${destFile}.roles_from`);
//   dumpRoles(dburl_target, `${destFile}.roles_target`);
//   apgdiffToFile(`${destFile}.roles_target`,
//                 `${destFile}.roles_from`,
//                 `${destFile}`);
//   fs.unlinkSync(`${destFile}.roles_from`);
//   fs.unlinkSync(`${destFile}.roles_target`);
// }


const migraToFile = (dburl_from, dburl_target, destFile, includeRoles, includePrivileges, withFilesBeginning = [], withFilesEnd = []) => {
  let cmd = _env.MIGRA_CMD;
  let params = ['--unsafe'];

  if (includePrivileges) {
    params.push('--with-privileges');
  }

  params.push(dburl_from);
  params.push(dburl_target);
  let options = {};
  console.log(`Diffing databases using migra`);
  console.log(`Writing the result to ${destFile.replace(_env.APP_DIR, '')}`);
  let p = (0, _common.runCmd)(cmd, params, options, false, false, false, ['ignore', 'pipe', 'pipe']);

  if (p.stdout) {
    let roles = '';

    if (includeRoles) {
      //diffRoles(dburl_from, dburl_target, `${destFile}.roles`)
      dumpRoles(dburl_from, `${destFile}.roles`);
      roles = _fs.default.readFileSync(`${destFile}.roles`, 'utf-8');

      _fs.default.unlinkSync(`${destFile}.roles`);
    }

    _fs.default.writeFileSync(destFile, ['-- generated with subzero-cli (https://github.com/subzerocloud/subzero-cli)', 'BEGIN;', withFilesBeginning.map(f => `\\ir ${f}`).join("\n"), roles, p.stdout.toString(), withFilesEnd.map(f => `\\ir ${f}`).join("\n"), 'COMMIT;'].join("\n"));
  }

  if (p.stderr.toString()) console.log(p.stderr.toString());
}; // const sqlworkbenchFile = (dburl_from, dburl_target, destFile, includeRoles, includePrivileges) => {
//   let cmd = JAVA_CMD
//   let splitConnectionString = new RegExp('postgresql:\/\/([^:]+):([^@]+)@(([^\/]+)\/([^?]+))')
//   let m1 = dburl_from.match(splitConnectionString)
//   let m2 = dburl_target.match(splitConnectionString)
//   let driver = "/usr/local/lib/postgresql-42.2.18.jar"
//   let options = {}
//   const DBURL_FROM = `driverjar=${driver},username=${m1[1]},password=${m1[2]},url=jdbc:postgresql://${m1[3]}`
//   const DBURL_TARGET = `driverjar=${driver},username=${m2[1]},password=${m2[2]},url=jdbc:postgresql://${m2[3]}`
//   const DEST_FILE = destFile+'.xml'
//   let WBCMD=[
//     'WbSchemaDiff',
//     '-styleSheet=wbdiff2pg.xslt',
//     `-referenceConnection="${DBURL_FROM}"`,
//     `-targetConnection="${DBURL_TARGET}"`,
//     `-file=${DEST_FILE}`,
//     `-xsltOutput=${destFile}`,
//     '-referenceSchema=*',
//     '-targetSchema=*',
//     '-includeViews=true',
//     '-includeProcedures=true',
//     '-includeIndex=true',
//   ]
//   if(includePrivileges){
//     console.log(`\x1b[31msqlworkbench has only partial support for --with-privileges\x1b[0m`);
//     WBCMD.push('-includeTableGrants=true')
//   }
//   let sqlworkbench_command="-command='"+WBCMD.join(' ') +" ;'";
//   let params = ['-jar', SQLWORKBENCH_JAR_PATH, sqlworkbench_command]
//   console.log(params)
//   console.log(`Diffing databases using sqlworkbench`);
//   console.log(`Writing the result to ${destFile}`);
//   let p = runCmd(cmd, params, options, false, false, false, ['ignore', 'pipe', 'pipe'])
//   if(p.stdout.toString()){
//     let roles = '';
//     if(includeRoles){
//       diffRoles(dburl_from, dburl_target, `${destFile}.roles`)
//       roles = fs.readFileSync(`${destFile}.roles`, 'utf-8');
//       fs.unlinkSync(`${destFile}.roles`);
//     }
//     fs.writeFileSync(destFile, roles + "\n" + p.stdout.toString());
//   }
//   if(p.stderr.toString())
//     console.log(p.stderr.toString());
//   if(fs.existsSync(destFile+'.xml')){fs.unlinkSync(destFile+'.xml')}
// };


const getPgVersion = dbUri => {
  // see first if we have an env setting
  if (_env.PG_VERSION) {
    return _env.PG_VERSION;
  }

  const env = Object.create(process.env),
        result = (0, _common.runCmd)(_env.PSQL_CMD, ['--quiet', '--tuples-only', '-c', 'show server_version', dbUri], {
    env: env
  }, true, false).stdout.toString().trim();
  let pgVersion = '13.1';

  if (result.indexOf(' ') !== -1) {
    pgVersion = result.substr(0, result.indexOf(' '));
  }

  return pgVersion;
};

_commander.default.command('init').option("--debug", "Verbose output and leaves the temporary files (used to create the migration) in place").option("--db-docker-image <image>", "Docker image used for temp postgres").option("--dry-run", "Don not create migrations files").option("--diff-tool <diff-tool>", "Use apgdiff, migra, sqlworkbench for diffing").option("--with-roles", "Include ROLE create/drop statements").option("--with-privileges", "Include grant/revoke commands").option("--include-beginning <\"file1.sql, file2.sql, ...\">", "Include specific sql scripts at the beginning of the migration").option("--include-end <\"file1.sql, file2.sql, ...\">", "Include specific sql scripts at the end of the migration").description('Setup sqitch config and create the first migration').action(options => {
  (0, _common.checkIsAppDir)();
  let withFilesBeginning = options.includeBeginning;
  let withFilesEnd = options.includeEnd;

  if (!withFilesBeginning) {
    withFilesBeginning = _env.MIGRATION_INCLUDE_BEGINNING;
  }

  if (!withFilesEnd) {
    withFilesEnd = _env.MIGRATION_INCLUDE_END;
  }

  initMigrations(options.debug, options.dbDockerImage, options.dryRun, options.diffTool, options.withRoles, options.withPrivileges, withFilesBeginning, withFilesEnd);
});

_commander.default.command('add <name>').option("-n, --note <note>", "Add sqitch migration note").option("-d, --no-diff", "Add empty sqitch migration (no diff)").option("--diff-tool <diff-tool>", "Use apgdiff, migra, sqlworkbench for diffing").option("--with-privileges", "Include grant/revoke commands (experimental)").option("--include-beginning <\"file1.sql, file2.sql, ...\">", "Include specific sql scripts at the beginning of the migration").option("--include-end <\"file1.sql, file2.sql, ...\">", "Include specific sql scripts at the end of the migration").option("--dry-run", "Don not create migrations files, only output the diff").option("--debug", "Verbose output and leaves the temporary files (used to create the migration) in place").option("--db-uri <uri>", "Diff against a database schema (By default we diff src/ and migrations/deploy/ directories)").option("--db-docker-image <image>", "Docker image used for temp postgres").description('Adds a new sqitch migration').action((name, options) => {
  (0, _common.checkIsAppDir)();

  if (options.withPrivileges) {
    console.log(`\x1b[31mWARNING: --with-privileges is an experimental feature and the output is incomplete, do not rely on it\x1b[0m`);
  }

  let withFilesBeginning = options.includeBeginning;
  let withFilesEnd = options.includeEnd;

  if (!withFilesBeginning) {
    withFilesBeginning = _env.MIGRATION_INCLUDE_BEGINNING;
  }

  if (!withFilesEnd) {
    withFilesEnd = _env.MIGRATION_INCLUDE_END;
  }

  addMigration(name, options.note, options.diff, options.dryRun, options.debug, options.dbUri, options.dbDockerImage, options.diffTool, options.withPrivileges, withFilesBeginning, withFilesEnd);
});

_commander.default.command('sqitch').description('run embeded sqitch').allowUnknownOption().action(() => {
  const sqitchParamas = process.argv.slice(process.argv.indexOf('sqitch') + 1);
  (0, _common.runCmd)(_env.SQITCH_CMD, sqitchParamas, {
    cwd: process.cwd()
  });
}).on('--help', function () {
  (0, _common.runCmd)(_env.SQITCH_CMD, ['--help'], {
    cwd: process.cwd()
  });
});

_commander.default.command('deploy <url>').description('Deploy sqitch migrations to a production database, url must have the `db:pg://${user}:${pass}@${host}:${port}/${db}` format').action(url => {
  (0, _common.checkIsAppDir)();
  (0, _common.checkMigrationsInitiated)();
  (0, _common.sqitchDeploy)(url);
});

_commander.default.parse(process.argv);